@@ -20,6 +20,7 @@ mod array;
 mod misc;
 pub use array::Array;
 pub use misc::RangeArgument;
+use array::Index;
 
 
 unsafe fn new_array<A: Array>() -> A {
@@ -44,7 +45,7 @@ unsafe fn new_array<A: Array>() -> A {
 /// ArrayVec can be converted into a by value iterator.
 pub struct ArrayVec<A: Array> {
     xs: NoDrop<A>,
-    len: u8,
+    len: A::Index,
 }
 
 impl<A: Array> Drop for ArrayVec<A> {
@@ -72,7 +73,7 @@ impl<A: Array> ArrayVec<A> {
     /// ```
     pub fn new() -> ArrayVec<A> {
         unsafe {
-            ArrayVec { xs: NoDrop::new(new_array()), len: 0 }
+            ArrayVec { xs: NoDrop::new(new_array()), len: Index::zero() }
         }
     }
 
@@ -87,7 +88,12 @@ impl<A: Array> ArrayVec<A> {
     /// assert_eq!(array.len(), 2);
     /// ```
     #[inline]
-    pub fn len(&self) -> usize { self.len as usize }
+    pub fn len(&self) -> usize { self.len.to_usize() }
+
+    fn set_len(&mut self, length: usize) {
+        debug_assert!(length <= self.capacity());
+        self.len = Index::from(length);
+    }
 
     /// Return the capacity of the **ArrayVec**.
     ///
@@ -122,11 +128,11 @@ impl<A: Array> ArrayVec<A> {
     /// ```
     pub fn push(&mut self, element: A::Item) -> Option<A::Item> {
         if self.len() < A::capacity() {
+            let len = self.len();
             unsafe {
-                let len = self.len();
                 ptr::write(self.get_unchecked_mut(len), element);
             }
-            self.len += 1;
+            self.len = Index::from(len + 1);
             None
         } else {
             Some(element)
@@ -149,11 +155,11 @@ impl<A: Array> ArrayVec<A> {
     /// assert_eq!(array.pop(), None);
     /// ```
     pub fn pop(&mut self) -> Option<A::Item> {
-        if self.len == 0 {
+        if self.len() == 0 {
             return None
         }
         unsafe {
-            self.len -= 1;
+            self.len = Index::from(self.len() - 1);
             let len = self.len();
             Some(ptr::read(self.get_unchecked_mut(len)))
         }
@@ -185,7 +191,7 @@ impl<A: Array> ArrayVec<A> {
         self.pop()
     }
 
-    /// Remove the element at **index** and shift the following elements down.
+    /// Remove the element at **index** and shift down the following elements.
     ///
     /// Return **Some(** *element* **)** if the index is in bounds, else **None**.
     ///
@@ -208,6 +214,53 @@ impl<A: Array> ArrayVec<A> {
         }
     }
 
+    /// Insert **element** in position **index**.
+    ///
+    /// Shift up all elements after **index**. If any is pushed out, it is returned.
+    ///
+    /// Return None if no element is shifted out.
+    ///
+    /// ## Examples
+    /// ```
+    /// use arrayvec::ArrayVec;
+    ///
+    /// let mut array = ArrayVec::<[_; 2]>::new();
+    ///
+    /// assert_eq!(array.insert(0, "x"), None);
+    /// assert_eq!(array.insert(0, "y"), None);
+    /// assert_eq!(array.insert(0, "z"), Some("x"));
+    /// assert_eq!(array.insert(1, "w"), Some("y"));
+    /// assert_eq!(&array[..], &["z", "w"]);
+    ///
+    /// ```
+    pub fn insert(&mut self, index: usize, element: A::Item) -> Option<A::Item> {
+        if index >= self.capacity() {
+            return Some(element);
+        }
+        let mut ret = None;
+        let old_len = self.len();
+        if old_len == self.capacity() {
+            ret = self.remove(old_len - 1);
+        }
+        let len = self.len();
+
+        // follows is just like Vec<T>
+        unsafe { // infallible
+            // The spot to put the new value
+            {
+                let p = self.as_mut_ptr().offset(index as isize);
+                // Shift everything over to make space. (Duplicating the
+                // `index`th element into two consecutive places.)
+                ptr::copy(&*p, p.offset(1), len - index);
+                // Write it in, overwriting the first copy of the `index`th
+                // element.
+                ptr::write(&mut *p, element);
+            }
+            self.set_len(len + 1);
+        }
+        ret
+    }
+
     /// Create a draining iterator that removes the specified range in the vector
     /// and yields the removed items from start to end. The element range is
     /// removed even if the iterator is not consumed until the end.
@@ -249,7 +302,7 @@ impl<A: Array> ArrayVec<A> {
 
         unsafe {
             // set self.vec length's to start, to be safe in case Drain is leaked
-            self.len = start as u8;
+            self.len = Index::from(start);
             Drain {
                 tail_start: end,
                 tail_len: len - end,
@@ -292,7 +345,7 @@ impl<A: Array> DerefMut for ArrayVec<A> {
 /// ```
 impl<A: Array> From<A> for ArrayVec<A> {
     fn from(array: A) -> Self {
-        ArrayVec { xs: NoDrop::new(array), len: A::capacity() as u8 }
+        ArrayVec { xs: NoDrop::new(array), len: Index::from(A::capacity()) }
     }
 }
 
@@ -352,14 +405,14 @@ impl<A: Array> IntoIterator for ArrayVec<A> {
     type Item = A::Item;
     type IntoIter = IntoIter<A>;
     fn into_iter(self) -> IntoIter<A> {
-        IntoIter { index: 0, v: self, }
+        IntoIter { index: Index::zero(), v: self, }
     }
 }
 
 
 /// By-value iterator for **ArrayVec**.
 pub struct IntoIter<A: Array> {
-    index: u8,
+    index: A::Index,
     v: ArrayVec<A>,
 }
 
@@ -372,16 +425,17 @@ impl<A: Array> Iterator for IntoIter<A> {
             None
         } else {
             unsafe {
-                let ptr = self.v.get_unchecked_mut(self.index as usize);
+                let index = self.index.to_usize();
+                let ptr = self.v.get_unchecked_mut(index);
                 let elt = ptr::read(ptr);
-                self.index += 1;
+                self.index = Index::from(index + 1);
                 Some(elt)
             }
         }
     }
 
     fn size_hint(&self) -> (usize, Option<usize>) {
-        let len = self.v.len() - self.index as usize;
+        let len = self.v.len() - self.index.to_usize();
         (len, Some(len))
     }
 }
@@ -393,7 +447,7 @@ impl<A: Array> DoubleEndedIterator for IntoIter<A> {
             None
         } else {
             unsafe {
-                self.v.len -= 1;
+                self.v.len = Index::from(self.v.len() - 1);
                 let len = self.v.len();
                 let elt = ptr::read(self.v.get_unchecked_mut(len));
                 Some(elt)
@@ -408,7 +462,7 @@ impl<A: Array> Drop for IntoIter<A> {
     fn drop(&mut self) {
         // exhaust iterator and clear the vector
         while let Some(_) = self.next() { }
-        self.v.len = 0;
+        self.v.len = Index::zero();
     }
 }
 
@@ -480,7 +534,7 @@ impl<'a, A: Array> Drop for Drain<'a, A>
                 let src = source_vec.as_ptr().offset(tail as isize);
                 let dst = source_vec.as_mut_ptr().offset(start as isize);
                 ptr::copy(src, dst, self.tail_len);
-                source_vec.len = (start + self.tail_len) as u8;
+                source_vec.len = Index::from(start + self.tail_len);
             }
         }
     }
@@ -582,6 +636,17 @@ fn test_simple() {
     assert_eq!(sum_len, 8);
 }
 
+#[test]
+fn test_u16_index() {
+    const N: usize = 4096;
+    let mut vec: ArrayVec<[_; N]> = ArrayVec::new();
+    for _ in 0..N {
+        assert!(vec.push(1u8).is_none());
+    }
+    assert!(vec.push(0).is_some());
+    assert_eq!(vec.len(), N);
+}
+
 #[test]
 fn test_iter() {
     let mut iter = ArrayVec::from([1, 2, 3]).into_iter();
@@ -696,3 +761,19 @@ fn test_drain_oob() {
     v.pop();
     v.drain(0..8);
 }
+
+#[test]
+fn test_insert() {
+    let mut v = ArrayVec::from([]);
+    assert_eq!(v.push(1), Some(1));
+    assert_eq!(v.insert(0, 1), Some(1));
+
+    let mut v = ArrayVec::<[_; 3]>::new();
+    v.insert(0, 0);
+    v.insert(1, 1);
+    v.insert(2, 2);
+    v.insert(3, 3);
+    assert_eq!(&v[..], &[0, 1, 2]);
+    v.insert(1, 9);
+    assert_eq!(&v[..], &[0, 9, 1]);
+}
